from typing import Callable, Type

from .action import RecursiveAction
from .core import AbstractParam, Const, ParaO
from .output import Output
from .print import PPrint
from .run import RunAct, RunAction

pprint = PPrint()


class Task[R](ParaO):
    code_version: Const
    run: RunAction[R]
    output: Callable[[RunAct[R]], Output[R]] | Type[Output[R]] = Output

    def __init_subclass__(cls):
        v = cls.__dict__.get("code_version")
        if v is not None and not isinstance(v, AbstractParam):
            cls.code_version = Const(v)
        r = cls.__dict__.get("run")
        if r is not None and not isinstance(r, AbstractParam):
            cls.run = RunAction(r)
        return super().__init_subclass__()

    @RecursiveAction
    def remove(self, depth: int):
        out = self.run.output
        if out.exists:
            out.remove()
            after = "removed"
        else:
            after = "missing"
        pprint.pprint(self, indent=2 * depth, after=after)

    @RecursiveAction
    def status(self, depth: int):
        after = "done" if self.run.done else "missing"
        pprint.pprint(self, indent=2 * depth, after=after)

    @RecursiveAction
    def print(self, depth: int):
        pprint.pprint(self, indent=2 * depth)



from collections import defaultdict
from .core import AbstractParam, Fragment, KeyTE, ParaO, get_inner_parao
from warnings import warn


def _match_recusive(instance: ParaO, fragments: tuple[Fragment]):
    cls = instance.__class__
    op = cls.__own_parameters__

    com: list[Fragment] = []
    sub: defaultdict[AbstractParam, list[Fragment]] = defaultdict(com.copy)
    for frag in fragments:
        com.append(frag)
        for s in sub.values():
            s.append(frag)

        if (param := op.got(frag.param)) and frag.is_type_ok(cls):
            if frag.inner is None:
                yield instance, param
            else:
                sub[param].append(frag.inner)

    for param in op.values():
        frags = tuple(sub[param]) if param in sub else fragments
        for inner in get_inner_parao(param.__get__(instance)):
            yield from _match_recusive(inner, frags)


def find_value(self: ParaO, key: KeyTE):
    frag = Fragment.make(key, None)
    cand = _match_recusive(self, (frag,))

    try:
        inst0, param0 = next(cand)
    except StopIteration:
        raise NoSuchParameter(str(key))

    first = True
    for inst, param in cand:
        if param is param0 and param.significant and inst == inst0:
            continue
        if first:
            value0 = param0.__get__(inst0)
        if param.__get__(inst) == value0:
            if first:
                warn(str(key), MultipleSameValues, stacklevel=2)
                first = False
        else:
            raise InconsistentValues(str(key))
    if first:
        value0 = param0.__get__(inst0)

    return value0


class NoSuchParameter(AttributeError): ...


class InconsistentValues(ValueError): ...


class MultipleSameValues(RuntimeWarning): ...


class ItemSteno(ParaO):
    __getitem__ = find_value


class AttrSteno(ParaO):
    def __getattr__(self, name: str):
        if not name.startswith("_"):
            return find_value(self, name)
        return super().__getattr__(name)


class Steno(ItemSteno, AttrSteno): ...



import pytest

from parao.core import ParaO, Param, Prop
from parao.steno import (
    InconsistentValues,
    MultipleSameValues,
    NoSuchParameter,
    Steno,
)


class Layer1(ParaO):
    foo = Param[int](0)
    bar = Param[int](0)


class Layer2(ParaO):
    left = Param[Layer1]()
    right = Param[Layer1]()


class Layer3(Steno, ParaO):
    inner = Param[Layer2]()


class Chain(Steno, ParaO):
    depth = Param[int]()
    aux = Param[int]()

    @Prop
    def inner(self) -> "Chain | None":
        if self.depth > 0:
            return Chain(depth=self.depth - 1)


def test_different():
    inst = Layer3({"foo": 1, ("right", "foo"): 2})
    assert inst["left", "foo"] == 1

    assert inst.right is inst.inner.right
    with pytest.raises(InconsistentValues):
        inst.foo
    with pytest.raises(NoSuchParameter):
        inst.wrong
    with pytest.raises(AttributeError):
        inst._ignored


def test_same():
    with pytest.warns(MultipleSameValues):
        assert Layer3({"foo": 3, ("right", "bar"): 4}).foo == 3


def test_identical():
    assert Layer3({"foo": 3}).foo == 3
    assert Layer3({"foo": 3}).foo == 3


def test_deep():
    ch = Chain({"depth": 3, (Chain, Chain.aux): 1})
    with pytest.warns(MultipleSameValues):
        assert ch["inner", "inner", "aux"] == 1



from pprint import PrettyPrinter
from shutil import get_terminal_size

from .core import ParaO


class PPrint(PrettyPrinter):
    def __init__(self, *args, width: int = -1, **kwargs):
        if width < 0:  # pragma: no branch
            width += get_terminal_size().columns
        super().__init__(*args, width=width, **kwargs)

    def pprint(self, object, indent: int = 0, after: str = ""):
        if self._stream is not None:  # pragma: no branch
            assert indent >= 0
            if after:
                after = f": {after}"
            self._stream.write(indent * self._indent_per_level * " ")
            self._format(object, self._stream, indent, len(after), {}, 0)
            self._stream.write(f"{after}\n")

    def _pprint_parao(self, object: ParaO, stream, indent, allowance, context, level):
        pre = f"{object.__class__.__fullname__}("
        stream.write(pre)

        if level > 1:
            stream.write("...")
        else:
            items = [
                (name, value)
                for name, value, neutral in object.__rich_repr__()
                if value != neutral
            ]
            if self._sort_dicts:  # pragma: no branch
                items.sort()
            stream.write("\n " + " " * indent)
            self._format_namespace_items(
                items, stream, indent + 1, allowance + 1, context, level
            )
            stream.write("\n" + " " * indent)

        stream.write(")")

    _dispatch = PrettyPrinter._dispatch.copy()
    _dispatch[ParaO.__repr__] = _pprint_parao



from typing import Annotated, Any
from unittest import TestCase
from parao.cast import cast


class TestCasting(TestCase):
    def test_primitives(self):
        self.assertEqual(cast("123", int), 123)
        self.assertEqual(cast("1.3", float), 1.3)
        self.assertEqual(cast("1j", complex), 1j)
        self.assertEqual(cast(123, str), "123")
        self.assertRaises(TypeError, lambda: cast("y", bool))

        # None
        self.assertEqual(cast(None, None), None)
        self.assertEqual(cast(None, type(None)), None)
        self.assertRaises(TypeError, lambda: cast(True, None))

    def test_containers(self):

        self.assertEqual(cast(["123"], list[int]), [123])
        self.assertEqual(cast({"123"}, set[int]), {123})
        self.assertEqual(cast(frozenset({"123"}), frozenset[int]), frozenset({123}))

        self.assertEqual(cast({"123": 456}, dict[int, str]), {123: "456"})
        self.assertEqual(cast([("123", 456)], dict[int, str]), {123: "456"})

        # no str/bytes to sequence
        self.assertRaises(TypeError, lambda: cast("123", list[int]))
        self.assertRaises(TypeError, lambda: cast(b"123", tuple[int, ...]))

        # empty tuple
        self.assertEqual(cast([], tuple[()]), ())
        self.assertRaises(TypeError, lambda: cast([1], tuple[()]))

        # any tuple
        self.assertEqual(cast([1, 2, 3], tuple), (1, 2, 3))
        self.assertEqual(cast([1, 2, 3], tuple[Any, ...]), (1, 2, 3))
        self.assertRaises(TypeError, lambda: cast([1], tuple[...]))  # type: ignore

        # fixed tuple
        self.assertEqual(cast([1, 2, 3], tuple[int, str, float]), (1, "2", 3.0))
        self.assertRaises(TypeError, lambda: cast([], tuple[int]))

    def test_complex(self):
        self.assertRaises(ValueError, lambda: cast(1.2, int))
        self.assertEqual(cast("1.2", int | float), 1.2)
        self.assertRaises(TypeError, lambda: cast("foo", int | float))
        self.assertEqual(cast("123", Annotated[int, str]), 123)

        class Foo:
            @classmethod
            def __cast_from__(cls, value, original_type):
                raise NotImplementedError

        self.assertRaises(TypeError, lambda: cast(1, Foo))

        class Bar:
            def __cast_to__(self, typ, original_type):
                try:
                    return typ(1.2)
                except Exception:
                    raise NotImplementedError

        self.assertEqual(cast(Bar(), int), 1)
        self.assertRaises(TypeError, lambda: cast(Bar(), list))

        class Boo[T]: ...

        self.assertRaises(TypeError, lambda: cast(1, Boo[int]))



from unittest import TestCase
from parao.misc import ContextValue, PeekableIter, safe_len, safe_repr, is_subseq


class TestContextValue(TestCase):
    def test_defaults(self):

        cv = ContextValue("cv")
        sentinel = object()
        self.assertIs(cv(default=sentinel), sentinel)


class TestMisc(TestCase):
    def test_safe(self):

        class Foo:
            def __repr__(self):
                raise RuntimeError()

        self.assertRaises(RuntimeError, lambda: repr(Foo()))
        self.assertEqual(safe_repr(o := Foo()), object.__repr__(o))

        self.assertRaises(TypeError, lambda: len(Foo()))
        self.assertEqual(safe_len(Foo(), (o := object())), o)

    def test_peekable(self):

        tpl = object(), object(), object()

        pi = PeekableIter(tpl)

        self.assertIs(pi.peek(), tpl[0])
        self.assertIs(pi.peek(), tpl[0])
        self.assertEqual(tuple(pi), tpl)
        self.assertRaises(StopIteration, lambda: pi.peek())
        self.assertIs(pi.peek(o := object()), o)

    def test_is_subseq(self):
        self.assertTrue(is_subseq("india", "indonesia"))
        self.assertTrue(is_subseq("oman", "romania"))
        self.assertTrue(is_subseq("mali", "malawi"))
        self.assertFalse(is_subseq("mali", "banana"))
        self.assertFalse(is_subseq("ais", "indonesia"))
        self.assertFalse(is_subseq("ca", "abc"))



from timeit import timeit as _timeit


def bti(
    subject,
    r=5,
    n: int = None,
    n0: int = 100,
    nf: int = 10,
    tt: float = 0.3,
    p: int = 4,
):

    vs = []
    for i in range(r):
        v = _timeit(subject, number=n or n0)
        if not i and n is None:
            n = n0
            while v < tt:
                n *= nf
                v = _timeit(subject, number=n)
        vs.append(v)

    vn = len(vs)
    vm = sum(vs) / vn
    vv = sum(v * v for v in vs) / vn
    vd = (vv - vm * vm) ** 0.5
    vr = vd / vm
    print(f"{vm:.{p}f}+-{vd:.{p}f} ({100*vr:5.3f}%)")



from functools import cached_property
from typing import Callable, Iterable, Self, Type

from .action import BaseRecursiveAction, RecursiveAct
from .cast import Opaque
from .core import ParaO


class PseudoOutput(Opaque):
    """Special type that must only be handled directly lest it cause an error."""


class BaseOutput[T]:
    __slots__ = ("act",)

    def __init__(self, act: "RunAct[T]"):
        self.act = act

    @property
    def exists(self) -> bool:
        raise NotImplementedError  # pragma: no cover

    def load(self) -> T:
        raise NotImplementedError  # pragma: no cover

    def dump(self, data: T) -> T:
        raise NotImplementedError  # pragma: no cover

    def remove(self, missing_ok: bool = False) -> None:
        raise NotImplementedError  # pragma: no cover


class RunAct[T](RecursiveAct["RunAction[T]"]):
    __call__: Callable[[], T]

    def _func(self, sub: Iterable[Self], depth, outer) -> T:
        if self.output.exists:
            return self.output.load()
        else:  # TODO: here we need to shim in Pool.map/asnyc
            for s in sub:
                if not s.done:
                    s(depth=depth + 1, outer=outer)

            return self.output.dump(self.action.func(self.instance))

    @cached_property
    def output(self) -> BaseOutput[T]:
        output = self.action.output or self.instance.output
        if not (isinstance(output, type) and issubclass(output, BaseOutput)):
            raise TypeError(f"{output=} must by a BaseOutput subclass")
        return output(self)

    @property
    def done(self):
        return self.output.exists


class RunAction[R](BaseRecursiveAction[R, []]):
    _act: Type[RunAct[R]] = RunAct
    func: Callable[[ParaO], R]
    __get__: Callable[..., RunAct[R]]
    output: Type[BaseOutput] | None = None



from functools import _lru_cache_wrapper
from .core import _solve_name, _get_type_hints
from .action import _method_1st_arg
from .shash import _code

caches: list[_lru_cache_wrapper] = [
    _solve_name,
    _get_type_hints,
    _method_1st_arg,
    _code,
]


def clear():
    for cache in caches:
        cache.cache_clear()



from operator import attrgetter
import pickle
from unittest import TestCase
from unittest.mock import Mock
from warnings import catch_warnings
from parao.core import (
    UNSET,
    OwnParameters,
    AbstractParam,
    Arguments,
    Const,
    Expansion,
    Fragment,
    ParaO,
    Param,
    MissingParameterValue,
    Prop,
    TypedAlias,
    Unset,
    UntypedParameter,
    Value,
    eager,
    DuplicateParameter,
    ExpansionGeneratedKeyMissingParameter,
)
import pytest

uniq_object = object()


def test_unset():
    assert isinstance(UNSET, Unset)
    assert UNSET is Unset()
    assert Unset() is Unset()


def test_Value():
    val = Value(uniq_object)
    assert val.val is uniq_object
    assert val.prio == 0
    assert val.position is None

    assert repr(Value(None, prio=1, position=1)) == "Value(None, 1, 1)"


def test_Fragment():
    key = ("foo", "bar")
    f = Fragment.make(key, Value(uniq_object))
    assert f.param == key[0]
    assert f.types is None
    assert isinstance(f.inner, Fragment)
    i = f.inner
    assert i.param == key[1]
    assert i.types is None
    assert isinstance(i.inner, Value)

    assert (
        repr(Fragment.make(key, Value(None)))
        == "Fragment('foo', None, Fragment('bar', None, Value(None)))"
    )


def test_Arguments():
    tpl = (1, "foo", uniq_object)  # the are actually bad types ...
    assert Arguments(tpl) == tpl

    assert (
        Arguments(
            [
                Fragment.make(("foo",), Value(uniq_object, 123)),
                Fragment.make(("foo", "bar"), Value(uniq_object, 123)),
            ]
        )
        == Arguments.from_dict(
            {"foo": uniq_object, ("foo", "bar"): uniq_object},
            prio=123,
        )
        == Arguments.from_dict(
            [
                ("foo", uniq_object),
                (("foo", "bar"), uniq_object),
            ],
            prio=123,
        )
    )

    assert Arguments.from_list([]) is Arguments.EMPTY
    assert Arguments.from_list([a := Arguments()]) is a
    assert Arguments.from_list(
        [Fragment.make(("foo",), Value(uniq_object))]
    ) == Arguments([Fragment.make(("foo",), Value(uniq_object))])

    with pytest.raises(TypeError):
        Arguments(123)

    assert repr(Arguments.make({})) == "Arguments()"
    assert (
        repr(Arguments.make(key=123)) == "Arguments(Fragment('key', None, Value(123)),)"
    )
    assert (
        repr(Arguments.make(foo=123, bar=456))
        == "Arguments(Fragment('foo', None, Value(123)), Fragment('bar', None, Value(456)))"
    )


class TestParam(TestCase):
    def test_param(self):
        self.assertIs(Param(type=(o := object())).type, o)
        self.assertIs(Param[o := object()]().type, o)
        self.assertRaises(TypeError, lambda: Param[int, str])

        # missing name - not really triggerable by user
        self.assertIs(Param()._name(int), None)

    def test_typed_alias(self):
        with self.assertWarns(TypedAlias.TypedAliasMismatch):

            class WonkyParam[A, B, C](AbstractParam[B]): ...

        WonkyParam[int, str, bool]()

        class Sentinel:
            pass

        class StrangeParam(AbstractParam):
            type = Sentinel

        self.assertIs(StrangeParam().type, Sentinel)

        with self.assertWarns(TypedAlias.TypedAliasRedefined):

            class RedundantParam[T](AbstractParam[T]):
                TypedAlias.register(T, TypedAlias._typevar2name[T])

        with self.assertRaises(TypedAlias.TypedAliasClash):

            class ClashingParam[T](AbstractParam[T]):
                TypedAlias.register(T, "not" + TypedAlias._typevar2name[T])

        with self.assertWarns(TypedAlias.TypedAliasMismatch):

            class MismatchParam[R](AbstractParam[R]): ...

    def test_specialized(self):
        uniq_const = object()
        uniq_aux = object()
        uniq_return = object()
        uniq_override = object()

        class Special(ParaO):
            const = Const(uniq_const)

            prop: object

            @Prop(aux=uniq_aux)
            def prop(self):
                return uniq_return

        with self.assertRaises(AttributeError):
            Special.prop._on_prop_attr
        with self.assertRaises(AttributeError):
            Special.prop.on_func_attr

        Special.prop.func.on_func_attr = attr = object()
        self.assertIs(Special.prop.on_func_attr, attr)

        self.assertIs(Special(const=None).const, uniq_const)
        self.assertIs(Special.prop.aux, uniq_aux)
        self.assertIs(Special().prop, uniq_return)
        self.assertIs(Special(prop=uniq_override).prop, uniq_override)


class TestParaO(TestCase):
    def test_create(self):
        ParaO()

        class Sub(ParaO): ...

        self.assertIsInstance(Sub(), Sub)
        self.assertIsInstance(ParaO({ParaO: Sub}), Sub)
        self.assertIsInstance(ParaO({"__class__": Sub}), Sub)

        # cover some rare branches
        self.assertIsInstance(
            ParaO(
                Arguments(
                    (
                        Arguments.from_dict({ParaO: UNSET}),
                        Arguments.EMPTY,
                        Arguments.from_dict({ParaO: Sub}),
                    )
                )
            ),
            Sub,
        )

        self.assertRaises(TypeError, lambda: ParaO({ParaO: 123}))

        with (
            catch_warnings(action="ignore", category=OwnParameters.CacheReset),
            self.assertRaises(DuplicateParameter),
        ):
            Sub.foo1 = Sub.foo2 = Param()

        self.assertEqual(
            Sub().__repr__(compact="???"),
            "tests.test_core:TestParaO.test_create.<locals>.Sub(???)",
        )

    def test_own_params(self):
        class Sub(ParaO):
            foo: int = Param()
            bar: str = Param()

        self.assertEqual(Sub.__own_parameters__, {"foo": Sub.foo, "bar": Sub.bar})

        with self.assertWarns(OwnParameters.CacheReset):
            Sub.boo = Param(type=float)

        self.assertEqual(Sub.__own_parameters__["boo"], Sub.boo)

        with self.assertWarns(OwnParameters.CacheReset):
            Sub.boo = Param(type=complex)

        self.assertEqual(Sub.__own_parameters__["boo"], Sub.boo)

        with self.assertWarns(OwnParameters.CacheReset):
            del Sub.foo, Sub.bar, Sub.boo

        self.assertEqual(Sub.__own_parameters__, {})

        with self.assertWarns(OwnParameters.CacheReset):
            Sub.boo = None
        del Sub.boo

        Sub.__dunder__ = None
        del Sub.__dunder__

    def test_resolution_simple(self):
        class Sub(ParaO):
            foo: int = Param()
            bar = Param(None, type=str)
            boo = Param[bool]()
            notyp = Param(None)

        self.assertEqual(Sub.boo.type, bool)

        with self.assertRaises(MissingParameterValue):
            Sub().foo
        with self.assertWarns(UntypedParameter):
            Sub().notyp

        self.assertEqual(Sub({"foo": 123}).foo, 123)
        self.assertEqual(Sub({Sub.foo: 123}).foo, 123)
        self.assertEqual(Sub({(Sub, "foo"): 123}).foo, 123)
        self.assertEqual(Sub({(Sub, Sub, "foo"): 123}).foo, 123)
        self.assertEqual(Sub({(Sub, Sub.foo): 123}).foo, 123)

        self.assertEqual(Sub().bar, None)
        self.assertEqual(Sub(bar=123).bar, "123")

    def test_resolution_complex(self):
        class Sub(ParaO):
            foo: int = Param()
            bar: str = Param(None)

        class Sub2(Sub):
            boo: bool = Param()

        self.assertEqual(Sub({Sub: Sub2, "boo": True}).boo, True)

        class Wrap(ParaO):
            one: Sub = Param()
            other: Sub2 = Param()

        class More(ParaO):
            inner: Wrap = Param()

        for addr in [("one", "bar"), (Wrap.one, Sub.bar), (Wrap.one, Sub, "bar")]:
            with self.subTest(addr=addr):
                self.assertEqual(Wrap({addr: 123}).one.bar, "123")
                self.assertEqual(Wrap({addr: 123}).other.bar, None)

        # providing a dict
        self.assertEqual(Wrap(one=dict(foo=123)).one.foo, 123)

        # unsing instance's args
        self.assertEqual(Wrap(one=Sub2(foo=123)).one.foo, 123)
        self.assertEqual(Wrap(one=Sub2(foo=123).__args__).one.foo, 123)

        # direct instance providing
        self.assertEqual(Wrap(one=Sub(foo=123)).one.foo, 123)
        self.assertIs(Wrap(one=(s := Sub())).one, s)

        # self.assertEqual(More().inner)

        # obj = Wrap({(Sub, "foo"): 123})
        # self.assertEqual(obj.one.foo, 123)
        # self.assertEqual(obj.other.foo, 123)

    def test_common_base(self):
        class Base(ParaO):
            foo = Param[int](0)

        class Ext1(Base):
            pass

        class Ext2(Base):
            ext1 = Param[Ext1]()

        ext2 = Ext2(foo=1)
        self.assertEqual(ext2.foo, 1)
        self.assertEqual(ext2.ext1.foo, 0)

    def test_non_eager_parameter(self):
        class Foo(ParaO):
            bar = Param[int](eager=False)

        with eager(True):
            foo = Foo()
        with self.assertRaises(MissingParameterValue):
            foo.bar

    def test_expansion(self):
        class Foo(ParaO):
            bar = Param[int]()

        with eager(False):
            f = Foo(bar=[1, 2, 3])
            # raises on access
            self.assertRaises(Expansion, lambda: f.bar)

        with eager(True):
            self.assertRaises(Expansion, lambda: Foo(bar=[1, 2, 3]))
            try:
                Foo(Arguments.from_dict({"unused": 1}), bar=[1, 2, 3])
            except Expansion as exp:
                self.assertEqual(exp.param, Foo.bar)
                self.assertEqual(exp.param_name, "bar")
                self.assertEqual(exp.values, (1, 2, 3))

        self.assertEqual(repr(Expansion([1, 2, 3])), "Expansion(<3 values>)")

    def test_collect(self):
        class Foo(ParaO):
            bar = Param[int]()

        # function based
        func = Mock(return_value=True)

        class Wrap(ParaO):
            foo = Param[Foo](collect=func)

        with eager(True):
            inst = Wrap(bar=[1, 2, 3])
        exp = inst.foo
        func.assert_called_once_with(exp, inst)
        self.assertIsInstance(exp.source, Foo)
        self.assertIsInstance(exp, Expansion)
        # self.assertEqual(exp._unwind, [])
        self.assertEqual(exp.make_key(), ("bar",))
        self.assertEqual(exp.make_key(False), (Foo, "bar"))
        self.assertEqual(exp.make_key(False, use_cls=False), ("bar",))
        self.assertEqual(exp.make_key(False, use_name=False), (Foo, Foo.bar))
        with self.assertWarns(ExpansionGeneratedKeyMissingParameter):
            self.assertEqual(
                exp.make_key(False, use_param=False),
                (Foo, "bar"),
            )
        with self.assertWarns(ExpansionGeneratedKeyMissingParameter):
            self.assertEqual(
                exp.make_key(False, want=(Foo,), use_name=False), (Foo, Foo.bar)
            )
        self.assertEqual(exp.make_key(False, want=(Foo.bar,)), ("bar",))
        self.assertIsInstance(repr(exp), str)

        # bare argument based
        items = [[Foo], [Foo.bar], ["bar"]]
        for coll in items + [[it] for it in items]:
            with self.subTest(coll=coll), eager(True):
                Wrap.foo.collect = coll
                self.assertIsInstance(Wrap(bar=[1, 2, 3]).foo, Expansion)

    def test_expand(self):
        # uses two level expandable scenario

        class Foo(ParaO):
            bar = Param[int]()

        class Mid(ParaO):
            boo = Param[int](0)
            foo = Param[Foo]()

        class Wrap2(ParaO):
            mid = Param[Mid](collect=Mock(return_value=True))

        with eager(True):
            self.assertEqual(Wrap2(bar=[1, 2, 3]).mid.make_key(), ("bar",))
            self.assertEqual(
                Wrap2({("foo", "bar"): [1, 2, 3]}).mid.make_key(), ("foo", "bar")
            )
            self.assertEqual(
                Wrap2(foo=dict(bar=[1, 2, 3])).mid.make_key(), (Mid, "foo", "bar")
            )
            self.assertSequenceEqual(
                list(map(attrgetter("foo.bar"), Wrap2(bar=[1, 2, 3]).mid.expand())),
                [1, 2, 3],
            )
            self.assertSequenceEqual(
                list(
                    map(
                        attrgetter("foo.bar"),
                        Wrap2({("foo", "bar"): [1, 2, 3]}).mid.expand(),
                    )
                ),
                [1, 2, 3],
            )
            self.assertSequenceEqual(
                list(
                    map(
                        attrgetter("foo.bar"),
                        Wrap2({("mid", "foo", "bar"): [1, 2, 3]}).mid.expand(),
                    )
                ),
                [1, 2, 3],
            )
            self.assertSequenceEqual(
                list(
                    map(
                        attrgetter("boo", "foo.bar"),
                        Wrap2(boo=[1, -1], bar=[1, 2, 3]).mid.expand(),
                    )
                ),
                [
                    (1, 1),
                    (1, 2),
                    (1, 3),
                    (-1, 1),
                    (-1, 2),
                    (-1, 3),
                ],
            )

    def test_inner(self):
        out1 = Out()
        self.assertEqual(
            tuple(out1.__inner__), (out1.in1, *out1.in2, out1.in3u, out1.in3u)
        )

        out2 = Out(in2=[In(), In()])
        self.assertEqual(
            tuple(out2.__inner__), (out2.in1, *out2.in2, out2.in3u, out2.in3u)
        )

        with eager(True):
            out3 = Out({("in1", "exp"): [1, 2]})
        inner = tuple(out3.__inner__)
        self.assertEqual(inner[0].exp, 1)
        self.assertEqual(inner[1].exp, 2)
        self.assertEqual(inner[2:], (out3.in3u, out3.in3u))

    def test_pickle(self):
        pre = Out({(In, In.exp): 1, "in2": [In(exp=2, uniq=3)], "uniq": -1})
        post = pickle.loads(pickle.dumps(pre))
        self.assertEqual(pre, post)
        with self.assertRaises(pickle.PicklingError):
            pickle.dumps(bare_param)


class In(ParaO):
    exp = Param[int](0)

    @Prop
    def uniq(self) -> int:
        return id(self)


class Out(ParaO):
    in1 = Param[In](collect=[In.exp])
    in2 = Param[list[In]]([])

    in3u = Param[In](significant=False)

    @Prop
    def in3(self) -> dict:
        return {
            "deep": [
                "nested",
                ("structure", {"with", frozenset({"some", (self.in3u,) * 2})}),
            ]
        }


bare_param = Param[str]()



from contextlib import contextmanager
from functools import lru_cache
from inspect import Parameter, signature
from operator import attrgetter
from typing import Any, Callable, Concatenate, Iterable, Self, Type

from .core import UNSET, AbstractDecoParam, ParaO, TypedAlias, Unset, Value, eager
from .misc import ContextValue

__all__ = ["SimpleAction", "ValueAction", "RecursiveAction"]


@lru_cache
def _method_1st_arg_annotation[T](
    func: Callable[Concatenate[Any, T, ...], Any],
) -> Type[T] | Unset:
    for i, param in enumerate(signature(func).parameters.values()):
        if i == 1:
            if param.kind in (
                Parameter.POSITIONAL_ONLY,
                Parameter.POSITIONAL_OR_KEYWORD,
                Parameter.VAR_POSITIONAL,
            ):
                if param.annotation is Parameter.empty:
                    return UNSET
                else:
                    return param.annotation
            break
    return UNSET


class BaseAct[T, A: "BaseAction"]:
    __slots__ = ("action", "instance", "value")

    def __init__(self, action: A, instance: ParaO, value: T, position: int = 0):
        self.action = action
        self.instance = instance
        self.value = value
        self.position = position
        self._add()

    def _add(self):
        if self.value is not UNSET:
            Plan.add(self)

    @property
    def name(self) -> str:
        return self.action._name(self.instance.__class__)

    __call__: Callable


class BaseAction[T, R, **Ps](AbstractDecoParam[T, Callable[Concatenate[ParaO, Ps], R]]):
    significant = False
    _act: Type[BaseAct] = BaseAct

    TypedAlias.register(R, "return_type")

    def _type(self, cls, name):
        return self.type

    def _get(self, val, name, instance) -> BaseAct:
        pos = val.position if isinstance(val, Value) else 0
        val = super()._get(val, name, instance)
        return self._act(self, instance, val, pos)

    def _collect(self, expansion, instance):  # can't collect
        return False  # pragma: no cover

    __get__: Callable[..., BaseAct[T, Self]]


# simple variant
class SimpleAct[R](BaseAct[bool, "SimpleAction[R]"]):
    def _add(self):
        if self.value:
            Plan.add(self)

    def __call__(self) -> R:
        return self.action.func(self.instance)


class SimpleAction[R](BaseAction[bool, R, []]):
    _act = SimpleAct
    func: Callable[[ParaO], R]
    __get__: Callable[..., SimpleAct[R]]
    type = bool


# value variant
class ValueAct[T, R](BaseAct[T, "ValueAction[T, R]"]):
    def __call__(self, override: T | Unset = UNSET) -> R:
        value = self.value if override is UNSET else override
        if value is UNSET:
            return self.action.func(self.instance)
        else:
            return self.action.func(self.instance, value)


class ValueAction[T, R](BaseAction[T, R, [T]]):
    def _type(self, cls, name):
        typ = self.type
        if typ is UNSET:
            typ = _method_1st_arg_annotation(self.func)
        return typ

    _act = ValueAct
    func: Callable[[ParaO, T], R]
    __get__: Callable[..., ValueAct[T, R]]
    type: Type[T]


# recursive variant
class RecursiveAct[A: "RecursiveAction"](BaseAct[int | bool | None, A]):
    def _inner(self):
        name = self.name
        cls = self.action.__class__
        for inner in self.instance.__inner__:
            if other := inner.__class__.__own_parameters__.get(name):
                if isinstance(other, cls):
                    yield getattr(inner, name)

    def _func(self, sub: Iterable[Self], depth: int, outer: int):
        if not self.action.func(self.instance, depth):
            for s in sub:
                s(depth=depth + 1, outer=outer)

    def __call__(
        self,
        override: int | bool | None = None,
        *,
        depth: int = 0,
        outer: int = True,
    ):
        val = self.value if override is None else override
        if val is UNSET or val is None:
            val = outer
        elif val is False or val < 0:
            return

        return self._func(
            self._inner() if val else (),
            depth,
            val is True or val < 1 or val - 1,
        )


class BaseRecursiveAction[R, **Ps](BaseAction[int | bool | None, R, Ps]):
    type = int | bool | None


class RecursiveAction(BaseRecursiveAction[bool, [int]]):
    func: Callable[[ParaO, int], bool]
    __get__: Callable[..., RecursiveAct]
    _act = RecursiveAct


class Plan(list[BaseAct]):
    current = ContextValue["Plan"]("currentPlan", default=None)

    @classmethod
    def add(cls, act: BaseAct):
        if (curr := cls.current()) is not None:
            curr.append(act)

    @contextmanager
    def use(self, /, run: bool = False):
        with self.current(self), eager(True):
            yield
            if run:
                self.run()

    def run(self):
        self.sort(key=attrgetter("position"))
        while self:
            self.pop(0)()



from binascii import b2a_hex
from collections import OrderedDict
import sys
from types import ModuleType
from unittest import TestCase
from unittest.mock import Mock, patch

from parao.core import ParaO, Param
from parao.shash import (
    _SHash,
    UnsupportedError,
    UsesLocalsError,
    UsesMainWarning,
    UsesPickleWarning,
    bin_hash,
    hex_hash,
)


def fake_main_func(): ...


fake_main_func.__module__ = "__main__"


class Unpickable:
    __reduce__ = None


class ndarray:
    def tolist(self): ...


class custom_int(int): ...


class Custom(ParaO):
    foo = Param[int](1)
    bar = Param[str](neutral="old")
    ignored = Param(significant=False)


class TestShash(TestCase):
    def test_values(self):
        list_in_self1 = []
        list_in_self1.append(list_in_self1)
        list_in_self2 = []
        list_in_self2.append([list_in_self2])

        # fmt: off
        self.assertEqual(hex_hash(1), b"6b86b273ff34fce19d6b804eff5a3f5747ada4eaa22f1d49c01e52ddb7875b4b")
        self.assertEqual(hex_hash(1.234), b"00a4e3d579394c0826980361ff3ca00e4915d0dc33b3bcb8775453aed9e8b16d")
        self.assertEqual(hex_hash(5.678j), b"2852241032ed1e4fba97014e64f3e1935f2a9eaa2093755590f851583bfb0377")
        self.assertEqual(hex_hash("foo"), b"af693da41462ec24753ce51403bd1ee5b500cd6941f62f3c97456dd8ffa7f952")
        self.assertEqual(hex_hash(b"bar"), b"0f0cd6bc85f1855d43efd8f861ecd52a7a9360b7cd32537c85118e24d657abae")
        self.assertEqual(hex_hash(None), b"dc937b59892604f5a86ac96936cd7ff09e25f18ae6b758e8014a24c7fa039e91")
        self.assertEqual(hex_hash(Ellipsis), b"4637e99b28a1ce112e8f4009dbf144f3a75a04d3e4b0c30b084aef21c5e3cfe6")
        self.assertEqual(hex_hash((1, 2, 3)), b"42f483f79545c1746890398387b591f019c11d69c6d86c2aad1375c70ce63b43")
        self.assertEqual(hex_hash([1, 2, 3]), b"7a65d251c6a3708230a39017f4ee7c7bf5036abe9002093d2ea337593dffa164")
        self.assertEqual(hex_hash({1, 2, 3}), b"058353b989d130371d502fd7730261695cd2e76e884e415ef00cbfb147d833c7")
        self.assertEqual(hex_hash(frozenset({1, 2, 3})), b"7f1cc479a8fa9c9ab41aaaa0804169febd62d7ee7ee674a29b87076efbe6e9f0")
        self.assertEqual(hex_hash({1: 2, 3: 4}), b"4b2b7e92da0e61c046a3d0a8626342bdb415b7acac138d33c18bb516877e1d66")
        self.assertEqual(hex_hash(OrderedDict({1: 2, 3: 4})), b"3d9cd99e702a126ca217205648b564959805edc71fe20a4a4d78abe265d2321a")
        self.assertEqual(hex_hash(OrderedDict({3: 4, 1: 2})), b"56305e4941ba58d828f7e406ea94c527ddb859c3a01caad9bed385d2dd0de944")
        self.assertEqual(hex_hash(range(1, 2, 3)), b"1cc18cce33f088254d3a7f0d04ef38f06de245547c27f19e8b51647f90b9a520")
        self.assertEqual(hex_hash(slice(1, 2, 3)), b"5a6c5731911d7c5bdedf4d9745d43ee2607a372650987c0448f273f087fd2be8")
        # wonky types
        self.assertEqual(hex_hash(sys), b"518b67e652531c5fe7e25d6b2c3b4ef6224e7d90da2091967dd47eb082b26a19")
        self.assertEqual(hex_hash(list_in_self1), b"16db3cdec82fa1efb8cd6bdcaf25bbc9cf35e5f9a11444d61ddc924b8f45df2d")
        self.assertEqual(hex_hash(list_in_self2), b"28f0ac587249fed6eae6d8e29cb25fec492f92979a52362b06fcddfdbd44f61e")
        # fmt: on

    def test_equivalence(self):
        self.assertEqual(b2a_hex(bin_hash(None)), hex_hash(None))

        self.assertEqual(hex_hash(set(range(100))), hex_hash(set(reversed(range(100)))))

        self.assertEqual(hex_hash({1: 2, 3: 4}), hex_hash({3: 4, 1: 2}))

        self.assertEqual(hex_hash(custom_int(5)), hex_hash(5))

        self.assertNotEqual(hex_hash([list, []]), hex_hash([[list]]))

    def test_difficulties(self):
        self.assertRaises(UnsupportedError, lambda: bin_hash(Unpickable()))

    def test_bad_unpickable(self):
        for probe in [
            object(),
            # from types
            len,
            [].append,
            object.__init__,
            object().__str__,
            str.join,
            # dict.__dict__["fromkeys"],
        ]:
            with self.subTest(probe=probe), self.assertWarns(UsesPickleWarning):
                bin_hash(probe)

    def test_locals(self):
        def make_closure():
            a = 1

            def func():
                return a  # pragma: no cover

            return func

        self.assertRaises(UsesLocalsError, lambda: bin_hash(make_closure()))

    def test_main(self):
        self.assertWarns(UsesMainWarning, lambda: bin_hash(fake_main_func))

    def test_numpy(self):
        numpy = Mock(spec=ModuleType)

        numpy.ndarray = ndarray
        numpy.dtype = Mock(spec=type)

        with patch.dict(sys.modules, numpy=numpy):
            del sys.modules["parao.shash"]

            from parao.shash import hex_hash

            nda = Mock(spec=ndarray)
            nda.tolist = Mock(return_value=[[[1], [2], [3]]])

            self.assertEqual(
                hex_hash(nda),
                b"38cbdfe4ca762f8eecf90ad461afe5b440ef1a89f525b87c450776ace64e5367",
            )

            nda.tolist.assert_called_once()

    def test_custom(self):
        custom = Mock()
        custom.__shash__ = Mock()
        custom.__shash__.return_value = b"foobar"

        s = _SHash()

        self.assertEqual(s(custom), b"foobar")
        custom.__shash__.assert_called_once_with(s)

    def test_parao(self):
        c = Custom(bar="")
        self.assertIs(bin_hash(c), bin_hash(c))
        self.assertEqual(repr(c), "tests.test_shash:Custom(bar='')")

        self.assertEqual(
            hex_hash(Custom(foo=1, bar="val")),
            b"e9259685825ceb325388dbc04057aa20c3abe7888cff403a63fb8ae92027a466",
        )
        self.assertEqual(
            hex_hash(Custom(bar="val")),
            b"e9259685825ceb325388dbc04057aa20c3abe7888cff403a63fb8ae92027a466",
        )
        self.assertEqual(
            hex_hash(Custom(bar="val", foo=2)),
            b"ef3ba52e2d4ddbe698e1adb3b8451bc8cf1e2157f63105af0c8ee3229c5f2ecb",
        )

        with patch.object(Custom.bar, "neutral", "old"):
            hash_old = hex_hash(Custom(bar="old"))
        with patch.object(Custom.bar, "significant", False):
            hash_ign = hex_hash(Custom())
        self.assertEqual(hash_old, hash_ign)



from ..core import ParaO

# alternative parameter ordering


class DefinitionOrder(ParaO):
    @classmethod
    def __dir__(self): ...


class CreationORder(ParaO):
    @classmethod
    def __dir__(self):
        orig = super().__dir__()



from .action import RecursiveAction, SimpleAction, ValueAction
from .cast import Opaque  # noqa: F401
from .cli import CLI
from .core import Const, Param, ParaO, Prop
from .output import JSON, Dir, File, Output, Pickle
from .run import RunAction
from .task import Task
from .steno import ItemSteno, AttrSteno, Steno

__all__ = [
    "ParaO",
    "Param",
    "Const",
    "Prop",
    "CLI",
    "SimpleAction",
    "ValueAction",
    "RecursiveAction",
    "Task",
    "RunAction",
    "Output",
    "File",
    "Dir",
    "JSON",
    "Pickle",
    "ItemSteno",
    "AttrSteno",
    "Steno",
]
